/*
 * Copyright (c) 1994, 1995, 1997, 1999, 2000, 2005, 2006, 2007, 2008
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] =
"@(#) Copyright (c) 1994, 1995, 1997, 1999, 2000, 2005, 2006, 2007, 2008\n\
	The Regents of the University of California.  All rights reserved.\n";
#endif /* not lint */

#ifndef lint
#if 0
static char sccsid[] = "@(#)tftpd.c	8.1 (Berkeley) 6/4/93";
#endif
static const char rcsid[] =
  "$FreeBSD: src/libexec/tftpd/tftpd.c,v 1.37 2005/05/31 17:22:53 maxim Exp $";
#endif /* not lint */

/*
 * Trivial file transfer protocol server.
 *
 * This version includes many modifications by Jim Guyton
 * <guyton@rand-unix>.
 *
 * Rewritten to handle multiple concurrent clients in a single process
 * by Craig Leres, leres@ee.lbl.gov.
 */

#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>		/* must come after sys/time.h */

#include <netinet/in.h>

#include <arpa/tftp.h>
#include <arpa/inet.h>

#include <ctype.h>
#include <errno.h>
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#ifdef HAVE_LIBUTIL_H
#include <libutil.h>
#endif
#include <memory.h>
#include <netdb.h>
#include <signal.h>
#include <pwd.h>
#include <setjmp.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>

#include "gnuc.h"
#ifdef HAVE_OS_PROTO_H
#include "os-proto.h"
#endif

#include "dhcpdconf.h"
#include "setsignal.h"
#include "tftpsubs.h"
#include "util.h"

#ifndef OACK
#define OACK		06		/* option acknowledgement */
#endif

#ifndef EOPTNEG
#define EOPTNEG		8		/* option negotiation failed */
#endif

/* XXX */
#undef LOG_DEBUG
#define LOG_DEBUG LOG_INFO

/* Everything we need to keep track of a client */

struct options {
	const char	*o_type;
	char	*o_request;
	int	o_reply;	/* turn into union if need be */
};

struct options protooptions[] = {
	{ "timeout",	NULL, 0 },	/* OPT_TIMEOUT */
	{ "blksize",	NULL, 0 },	/* OPT_BLKSIZE */
	{ "tsize",	NULL, 0 },	/* OPT_TSIZE */
	{ NULL,		NULL, 0 }
};
#define NUMOPTIONS (sizeof(protooptions) / sizeof(protooptions[0]))

enum opt_enum {
	OPT_TIMEOUT = 0,
	OPT_BLKSIZE = 1,
	OPT_TSIZE = 2,
};

struct client {
	struct sockaddr_in sin;		/* address and port of client */
	time_t wakeup;			/* time to wakeup */
	int retries;			/* retries left until we give up */
	char *buf;			/* pointer to received packet */
	struct tftphdr *dp;		/* pointer to output packet */
	int bufcc;			/* size of data in tp */
	int dpcc;			/* size of data in dp */
	int proc;			/* client proc */
	int state;			/* client state */
	int segsize;			/* segment size */
	FILE *file;			/* open file pointer */
	u_short block;			/* block in file */
	int convert;			/* convert to ascii if true */
	struct options options[NUMOPTIONS];
	struct ra ra;			/* read ahead buffer stuff */
};

/* Client states */
#define CP_FREE		0
#define CP_TFTP		1		/* ready to start a session */
#define CP_SENDFILE	2		/* ready to send a block */
#define CP_RECVFILE	3		/* ready to recv a block */
#define CP_OACK		4		/* sent options ack */
#define CP_DONE		5		/* all done */

#define ST_INIT		1		/* initialize */
#define ST_SEND		2		/* send a block */
#define ST_RECV		2		/* receive a block */
#define ST_WAITACK	3		/* waiting for a block to be ack'ed */
#define ST_LASTACK	4		/* sent last block, wait for last ack */

/* Internal statistics */
struct stats {
	u_int32_t retrans;		/* number of retransmissions */
	u_int32_t timeout;		/* number of timed out clients */
	u_int32_t badop;		/* number of was expecting ACK */
	u_int32_t errorop;		/* received ERROR from client */
	u_int32_t badfilename;		/* filename wasn't parsable */
	u_int32_t badfilemode;		/* file mode wasn't parsable */
	u_int32_t wrongblock;		/* client ACKed wrong block */
	u_int32_t fileioerror;		/* errors reading the file */
	u_int32_t sent;			/* number of packets sent */
	u_int32_t recv;			/* number of packets received */
	u_int32_t maxclients;		/* maximum concurrent clients */
};

#define	TIMEOUT		5
#define	MAX_TIMEOUTS	5

int	rexmtval = TIMEOUT;
int	max_rexmtval = 2*TIMEOUT;

#define WAIT_RETRIES 5			/* retransmits before giving up */

#define MORECLIENTS 10			/* add this many when we need more */

struct	sockaddr_in from;

/*
 * Null-terminated directory prefix list for absolute pathname requests and
 * search list for relative pathname requests.
 *
 * MAXDIRS should be at least as large as the number of arguments that
 * inetd allows (currently 20).
 */
#define MAXDIRS	20
static struct dirlist {
	const char	*name;
	int	len;
} dirs[MAXDIRS + 1];
static int	suppress_naks;
static int	logging;
static int	ipchroot;
static int	create_new = 0;
static mode_t	mask = S_IWGRP|S_IWOTH;

/* Forwards */
static	void ack(struct client *);
static	char *errtomsg(int);
static	struct client *findclient(struct sockaddr_in *);
static	void freeclient(struct client *);
static	void nak(struct client *, int, char const *, ...);
static	struct client *newclient(void);
static	void oack(struct client *);
static	void process(struct sockaddr_in *, char *, int);
static	void recvfile(struct client *);
static	void recvoack(struct client *);
static	void runclient(struct client *);
static	void runtimer(void);
static	void xmitfile(struct client *);
static	void tftp(struct client *);
static	int validate_access(struct client *, char **, int);
static	__dead void finish(int) __attribute__((volatile));
static	int goodclient(u_int32_t);
static	void reduce_path(char *);

void	die(int);
void	logstats(int);
int	main(int, char *[]);

/* Globals */
char	*errstr;			/* detail for DENIED syslog */
int	debug;

/* Locals */
static	int suppress_naks;
static	int s;				/* tftp socket */
static	char *dhcpdconf;
static	int havedhcpdconf;
static	time_t now;			/* updated after packet reception */

static	int dostats;			/* need to syslog statistics */
static	int candostats;			/* ok to syslog statistics */
static	struct stats stats;		/* tftpd statistics */
struct rlimit rlimit = {
	RLIM_INFINITY,
	RLIM_INFINITY
};

static	struct client **clientlist;
static	int clientlistsize;		/* size of clientlist array */
static	int clientlistcnt;		/* number in use */

int
main(int argc, char *argv[])
{
	socklen_t fromlen;
	int ch, n;
	struct sockaddr_in me;
	char *chroot_dir = NULL;
	struct passwd *nobody;
	const char *chuser = "nobody";
	char *ipaddr = NULL;

	fd_set fds, rfds;
	struct timeval tv;
	struct timeval *tvp;
	time_t last;
	const char *msg;
	struct servent *sp;
	char buf[PKTSIZE];

	tzset();				/* syslog in localtime */

	openlog("tftpd", LOG_PID | LOG_NDELAY, LOG_FTP);
	while ((ch = getopt(argc, argv, "a:cCdDlns:u:U:w")) != EOF) {
		switch (ch) {

		case 'a':
			ipaddr = optarg;
			break;

		case 'd':
			++debug;
			++logging;
			break;

		case 'D':
			dhcpdconf = optarg;
			break;

		case 'c':
			ipchroot = 1;
			break;

		case 'C':
			ipchroot = 2;
			break;

		case 'l':
			++logging;
			break;

		case 'n':
			++suppress_naks;
			break;

		case 's':
			chroot_dir = optarg;
			break;

		case 'u':
			chuser = optarg;
			break;

		case 'U':
			mask = strtol(optarg, NULL, 0);
			break;

		case 'w':
			++create_new;
			break;

		default:
			fprintf(stderr, "usage: tftpd [-cClnw]"
			    " [-a ipaddr] [-D dhcpdconf]"
			    " [-s directory] [-u user]\n\t[-U umask]"
			    " [directory ...]\n");
			exit(1);
		}
	}
	if (optind < argc) {
		struct dirlist *dirp;

		/* Get list of directory prefixes. Skip relative pathnames. */
		for (dirp = dirs; optind < argc && dirp < &dirs[MAXDIRS];
		     optind++) {
			if (argv[optind][0] == '/') {
				dirp->name = argv[optind];
				dirp->len  = strlen(dirp->name);
				dirp++;
			}
		}
	}
	if (chroot_dir) {
		dirs->name = "/";
		dirs->len = 1;
	}
	if (ipchroot > 0 && chroot_dir == NULL) {
		lg(LOG_ERR, "-c requires -s");
		exit(1);
	}

	umask(mask);

	/* Submerge if not debugging */
	if (!debug)
		daemon(0, 0);

	sp = getservbyname("tftp", "udp");
	if (sp == NULL) {
		lg(LOG_ERR, "tftp/udp: unknown service");
		finish(1);
	}

	/* Read the configs */
	if (dhcpdconf != NULL) {
		msg = setdhcpdconf(dhcpdconf, 1);
		if (msg == NULL)
			++havedhcpdconf;
		else
			lg(LOG_ERR, "setdhcpdconf: %s", msg);
	}

	/* Setup our socket */
	s = socket(AF_INET, SOCK_DGRAM, 0);
	if (s < 0) {
		lg(LOG_ERR, "socket: %m");
		finish(1);
	}

	memset(&me, 0, sizeof(me));
	me.sin_family = AF_INET;
	if (ipaddr != NULL) {
		n = inet_pton(AF_INET, ipaddr, &me.sin_addr);
		if (n < 0) {
			lg(LOG_ERR, "inet_pton(%s): %m", ipaddr);
			finish(1);
		}
		if (n == 0) {
			lg(LOG_ERR, "inet_pton(%s): bad address", ipaddr);
			finish(1);
		}
		if (n != 1) {
			lg(LOG_ERR, "inet_pton(%s): bad status", ipaddr);
			finish(1);
		}
	} else
		me.sin_addr.s_addr = INADDR_ANY;
	me.sin_port = sp->s_port;
#ifdef HAVE_SOCKADDR_SA_LEN
	me.sin_len = sizeof(me);
#endif
	if (bind(s, (struct sockaddr *)&me, sizeof(me)) < 0) {
		lg(LOG_ERR, "bind: %m");
		finish(1);
	}

	if (chroot_dir) {
#ifdef notdef
		if (ipchroot > 0) {
			char *tempchroot;
			struct stat sb;
			int statret = 0;
			struct sockaddr_in sin;
			char hbuf[NI_MAXHOST];

			memcpy(&sin, &from, from.sin_len);
#ifdef HAVE_IPV6
			unmappedaddr((struct sockaddr_in6 *)&sin);
#endif
			getnameinfo((struct sockaddr *)&sin, sin.sin_len,
				    hbuf, sizeof(hbuf), NULL, 0,
				    NI_NUMERICHOST);
			asprintf(&tempchroot, "%s/%s", chroot_dir, hbuf);
			if (ipchroot == 2)
				statret = stat(tempchroot, &sb);
			if (ipchroot == 1 ||
			    (statret == 0 && (sb.st_mode & S_IFDIR)))
				chroot_dir = tempchroot;
		}
#endif
		/* Must get this before chroot because /etc might go away */
		if ((nobody = getpwnam(chuser)) == NULL) {
			lg(LOG_ERR, "%s: no such user", chuser);
			exit(1);
		}
		if (chroot(chroot_dir)) {
			lg(LOG_ERR, "chroot: %s: %m", chroot_dir);
			exit(1);
		}
		chdir("/");
		setgroups(1, &nobody->pw_gid);
		setuid(nobody->pw_uid);
	}

	/* (Try to) max out open file limit */
	if (setrlimit(RLIMIT_NOFILE, &rlimit) < 0 && errno != EPERM)
		lg(LOG_ERR, "setrlimit: %m");

	/* Setup signal handlers */
	(void)setsignal(SIGHUP, die);
	(void)setsignal(SIGINT, die);
	(void)setsignal(SIGTERM, die);
	(void)setsignal(SIGUSR1, logstats);
	if (debug)
		(void)setsignal(SIGQUIT, logstats);

	FD_ZERO(&fds);
	FD_SET(s, &fds);
	tv.tv_sec = rexmtval;
	tv.tv_usec = 0;
	last = 0;
	lg(LOG_INFO, "nofiles %d", getdtablesize());
	for (;;) {
		rfds = fds;
		if (clientlistcnt > 0)
			tvp = &tv;
		else
			tvp = NULL;
		++candostats;
		if (select(s + 1, &rfds, NULL, NULL, tvp) < 0) {
			/* Don't choke when we get ptraced */
			if (errno == EINTR)
				continue;
			lg(LOG_ERR, "select: %m");
			finish(1);
		}
		candostats = 0;
		if (dostats)
			logstats(0);

		if (FD_ISSET(s, &rfds)) {
			/* Process a packet */
			fromlen = sizeof(from);
			n = recvfrom(s, buf, sizeof(buf), 0,
			    (struct sockaddr *)&from, &fromlen);
			if (n < 0) {
				lg(LOG_ERR, "recvfrom: %m");
				continue;
			}

			/* Update now */
			now = time(NULL);

			/* Process this packet */
			process(&from, buf, n);
		} else
			now = time(NULL);

		/* Run the timer list, no more than once a second */
		if (clientlistcnt > 0 && last != now) {
			runtimer();
			last = now;
		}
	}
}

static void
reduce_path(char *fn)
{
	char *slash, *ptr;

	/* Reduce all "/+./" to "/" (just in case we've got "/./../" later */
	while ((slash = strstr(fn, "/./")) != NULL) {
		for (ptr = slash; ptr > fn && ptr[-1] == '/'; ptr--)
			;
		slash += 2;
		while (*slash)
			*++ptr = *++slash;
	}

	/* Now reduce all "/something/+../" to "/" */
	while ((slash = strstr(fn, "/../")) != NULL) {
		if (slash == fn)
			break;
		for (ptr = slash; ptr > fn && ptr[-1] == '/'; ptr--)
			;
		for (ptr--; ptr >= fn; ptr--)
			if (*ptr == '/')
				break;
		if (ptr < fn)
			break;
		slash += 3;
		while (*slash)
			*++ptr = *++slash;
	}
}

struct formats {
	const char	*f_mode;
	int	(*f_validate)(struct client *, char **, int);
	int	f_convert;
} formats[] = {
	{ "netascii",	validate_access,	1 },
	{ "octet",	validate_access,	0 },
#ifdef notdef
	{ "mail",	validate_user,		1 },
#endif
	{ 0,		NULL,			0 }
};

/* Handle initial connection protocol  */
static void
tftp(struct client *cl)
{
	char *cp, *ep;
	int *ip;
	int i, first = 1, has_options = 0, ecode;
	struct formats *pf;
	char *filename, *mode, *option, *ccp;
	char fnbuf[PATH_MAX];
	struct tftphdr *tp;
	int isclient;

	mode = NULL;
	tp = (struct tftphdr *)cl->buf;
	tp->th_opcode = ntohs((u_short)tp->th_opcode);
	cp = tp->th_stuff;
again:
	while (cp < cl->buf + cl->bufcc) {
		if (*cp == '\0')
			break;
		cp++;
	}
	if (*cp != '\0') {
		++stats.badfilename;
		nak(cl, EBADOP, "filename missing EOS");
		return;
	}
	i = cp - tp->th_stuff;
	if (i >= sizeof(fnbuf)) {
		++stats.badfilename;
		nak(cl, EBADOP, "filename too long");
		return;
	}
	memcpy(fnbuf, tp->th_stuff, i);
	fnbuf[i] = '\0';
	reduce_path(fnbuf);
	filename = fnbuf;
	if (first) {
		mode = ++cp;
		first = 0;
		goto again;
	}
	for (cp = mode; *cp; cp++)
		if (isupper(*cp))
			*cp = tolower(*cp);
	for (pf = formats; pf->f_mode; pf++)
		if (strcmp(pf->f_mode, mode) == 0)
			break;
	if (pf->f_mode == 0) {
		++stats.badfilemode;
		nak(cl, EBADOP, "unknown mode");
		return;
	}
	cl->convert = pf->f_convert;
	while (++cp < cl->buf + cl->bufcc) {
		for (i = 2, ccp = cp; i > 0; ccp++) {
			if (ccp >= cl->buf + cl->bufcc) {
				/*
				 * Don't reject the request, just stop trying
				 * to parse the option and get on with it.
				 * Some Apple Open Firmware versions have
				 * trailing garbage on the end of otherwise
				 * valid requests.
				 */
				goto option_fail;
			} else if (*ccp == '\0')
				i--;
		}
		for (option = cp; *cp; cp++)
			if (isupper(*cp))
				*cp = tolower(*cp);
		for (i = 0; cl->options[i].o_type != NULL; i++)
			if (strcasecmp(option, cl->options[i].o_type) == 0) {
				cl->options[i].o_request = ++cp;
				has_options = 1;
			}
		cp = ccp-1;
	}

option_fail:
	/* RFC 2349 */
	if (cl->options[OPT_TIMEOUT].o_request) {
		ip = &cl->options[OPT_TIMEOUT].o_reply;
		*ip = strtoul(cp, &ep, 10);
		if (*ep != '\0' || *ip < 1 || *ip > 255) {
			nak(cl, EOPTNEG, "timeout out of range");
			return;
		}
		if (*ip > max_rexmtval)
			*ip = max_rexmtval;
		rexmtval = *ip;
lg(LOG_DEBUG, "tftp timeout set to %d", rexmtval);
	}

	/* RFC 2348 */
	if (cl->options[OPT_BLKSIZE].o_request) {
		ip = &cl->options[OPT_BLKSIZE].o_reply;
		*ip = strtoul(cp, &ep, 10);
		if (*ep != '\0' || *ip < 8 || *ip > MAXSEGSIZE) {
			nak(cl, EOPTNEG, "blksize out of range");
			return;
		}
		cl->segsize = *ip;
lg(LOG_DEBUG, "tftp blksize set to %d", cl->segsize);
	}

	isclient = goodclient(cl->sin.sin_addr.s_addr);
	errstr = NULL;
	if (!isclient) {
		errstr = "not tftp client";
		ecode = EACCESS;
	} else
		ecode = (*pf->f_validate)(cl, &filename, tp->th_opcode);

	if (has_options && ecode == 0)
		oack(cl);
	if (logging)
		lg(LOG_INFO, "%s: %s request for %s: %s",
			inet_ntoa(cl->sin.sin_addr),
			tp->th_opcode == WRQ ? "write" : "read",
			filename, errtomsg(ecode));
	if (ecode) {
		/*
		 * Avoid storms of naks to a RRQ broadcast for a relative
		 * bootfile pathname from a diskless Sun.
		 */

		lg(LOG_INFO, "%s %s \"%s\" DENIED (%s)",
			inet_ntoa(cl->sin.sin_addr),
			tp->th_opcode == WRQ ? "write" : "read",
			filename,
			errstr != NULL ? errstr : errtomsg(ecode));
		if (suppress_naks && *filename != '/' && ecode == ENOTFOUND)
			cl->proc = CP_DONE;
		else
			nak(cl, ecode, errstr);
		return;
	}

	cl->state = ST_INIT;
	cl->buf = NULL;
	cl->bufcc = 0;
	if (tp->th_opcode == WRQ) {
		cl->proc = CP_RECVFILE;
		recvfile(cl);
	} else {
		cl->proc = CP_SENDFILE;
		xmitfile(cl);
	}
}

/*
 * Validate file access.  Since we
 * have no uid or gid, for now require
 * file to exist and be publicly
 * readable/writable.
 * If we were invoked with arguments
 * from inetd then the file must also be
 * in one of the given directory prefixes.
 * Note also, full path name must be
 * given as we have no login directory.
 */
int
validate_access(struct client *cl, char **filep, int mode)
{
	struct stat stbuf;
	int	fd;
	struct dirlist *dirp;
	static char pathname[MAXPATHLEN];
	char *filename = *filep;

if (mode == 256) {
lg(LOG_DEBUG, "bad mode %d", mode);
	mode = RRQ;
}
	/*
	 * Prevent tricksters from getting around the directory restrictions
	 */
	if (strstr(filename, "/../")) {
		errstr = "\"..\" in path";
		return (EACCESS);
	}

	if (*filename == '/') {
		/*
		 * Allow the request if it's in one of the approved locations.
		 * Special case: check the null prefix ("/") by looking
		 * for length = 1 and relying on the arg. processing that
		 * it's a /.
		 */
		for (dirp = dirs; dirp->name != NULL; dirp++) {
			if (dirp->len == 1 ||
			    (!strncmp(filename, dirp->name, dirp->len) &&
			     filename[dirp->len] == '/'))
				    break;
		}
		/* If directory list is empty, allow access to any file */
		if (dirp->name == NULL && dirp != dirs) {
			errstr = "not on list";
			return (EACCESS);
		}
		if (stat(filename, &stbuf) < 0) {
			errstr = strerror(errno);
			return (errno == ENOENT ? ENOTFOUND : EACCESS);
		}
		if ((stbuf.st_mode & S_IFMT) != S_IFREG) {
			errstr = "not a regular file";
			return (ENOTFOUND);
		}
		if (mode == RRQ) {
			if ((stbuf.st_mode & S_IROTH) == 0) {
				errstr = "not readable";
				return (EACCESS);
			}
		} else {
			if ((stbuf.st_mode & S_IWOTH) == 0) {
				errstr = "not writable";
				return (EACCESS);
			}
		}
	} else {
		int err;

		/*
		 * Relative file name: search the approved locations for it.
		 * Don't allow write requests that avoid directory
		 * restrictions.
		 */

		if (!strncmp(filename, "../", 3)) {
			errstr = "\"..\" in path";
			return (EACCESS);
		}

		/*
		 * If the file exists in one of the directories and isn't
		 * readable, continue looking. However, change the error code
		 * to give an indication that the file exists.
		 */
		errstr = "not found";
		err = ENOTFOUND;
		for (dirp = dirs; dirp->name != NULL; dirp++) {
			if (strcmp(dirp->name, "/") == 0)
				snprintf(pathname, sizeof(pathname), "%s%s",
					dirp->name, filename);
			else
				snprintf(pathname, sizeof(pathname), "%s/%s",
					dirp->name, filename);
			if (stat(pathname, &stbuf) == 0 &&
			    (stbuf.st_mode & S_IFMT) == S_IFREG) {
				if ((stbuf.st_mode & S_IROTH) != 0) {
					break;
				}
				errstr = "not readable";
				err = EACCESS;
			}
		}
		if (dirp->name != NULL)
			*filep = filename = pathname;
		else if (mode == RRQ)
			return (err);
	}

	/* RFC 2349 */
	if (cl->options[OPT_TSIZE].o_request) {
		if (mode == RRQ)
			cl->options[OPT_TSIZE].o_reply = stbuf.st_size;
		else {
			/* XXX Allows writes of all sizes. */
			cl->options[OPT_TSIZE].o_reply =
				atoi(cl->options[OPT_TSIZE].o_request);
		}
	}
	if (mode == RRQ)
		fd = open(filename, O_RDONLY);
	else {
		if (create_new)
			fd = open(filename, O_WRONLY|O_TRUNC|O_CREAT, 0666);
		else
			fd = open(filename, O_WRONLY|O_TRUNC);
	}
	if (fd < 0) {
		errstr = strerror(errno);
		return (errno + 100);
	}
	cl->file = fdopen(fd, (mode == RRQ)? "r" : "w");
	if (cl->file == NULL) {
		errstr = strerror(errno);
		close(fd);
		return (errno + 100);
	}
	return (0);
}

/*
 * Send the requested file.
 */
static void
xmitfile(struct client *cl)
{
	int oerrno, size;
	struct tftphdr *tp;
	struct tftphdr *dp;
	int opcode, readblock;
	volatile u_short block;

	readblock = 0;
	if ((tp = (struct tftphdr *)cl->buf) != NULL) {
		/* Process an input packet */
		opcode = ntohs((u_short)tp->th_opcode);

		/* Give up if the other side wants to */
		if (opcode == ERROR) {
			++stats.errorop;
			lg(LOG_INFO, "%s.%d ERROR",
			    inet_ntoa(cl->sin.sin_addr), cl->sin.sin_port);
			cl->proc = CP_DONE;
			return;
		}

		/* If he didn't ACK, ignore packet */
		if (opcode != ACK) {
			/* XXX this probably can't happen */
			lg(LOG_DEBUG, "%s.%d xmitfile: expected opcode %d",
			    inet_ntoa(cl->sin.sin_addr), cl->sin.sin_port,
			    opcode);
			return;
		}

		if (cl->state == ST_WAITACK || cl->state == ST_LASTACK) {
			/* If he didn't ack the right block, re-synch */
			block = ntohs((u_short)tp->th_block);
			if (block != cl->block) {
				++stats.wrongblock;
				/* XXX todo */
				/* Re-synchronize with the other side */
				(void) synchnet(&cl->ra);
				cl->state = ST_SEND;
				return;
			}

			/* He ack'ed the last block we sent him */
			if (debug > 1)
				lg(LOG_DEBUG, "%s.%d < ack %d",
				    inet_ntoa(cl->sin.sin_addr),
				    cl->sin.sin_port,
				    cl->block);

			/* Check if we're done */
			if (cl->state == ST_LASTACK) {
				cl->proc = CP_DONE;
				return;
			}

			/* Setup to send the next block */
			++cl->block;
			++readblock;
			cl->state = ST_SEND;
			cl->wakeup = 0;
			cl->retries = WAIT_RETRIES;
		}
	}

	/* See if it's time to resend */
	if (cl->wakeup != 0 && cl->wakeup <= now &&
	    (cl->state == ST_WAITACK || cl->state == ST_LASTACK)) {
		++stats.retrans;
		lg(LOG_DEBUG, "%s.%d retrans",
		    inet_ntoa(cl->sin.sin_addr), cl->sin.sin_port);
		cl->state = ST_SEND;
	}

	/* Initialize some stuff the first time through */
	if (cl->state == ST_INIT) {
		cl->block = 1;
		++readblock;
		cl->state = ST_SEND;
		cl->wakeup = 0;
		cl->retries = WAIT_RETRIES;
		cl->dp = r_init(&cl->ra);
	}

	/* If our first time through or the last block was ack'ed, read */
	size = 0;
	if (readblock) {
		errno = 0;
		size = readit(&cl->ra, cl->file, &dp, cl->convert, cl->segsize);
		cl->dpcc = size;
		cl->dp = dp;

		/* Bail if we run into trouble */
		if (size < 0) {
			++stats.fileioerror;
			nak(cl, errno + 100, NULL);
			return;
		}

		cl->dpcc += 4;
		dp->th_opcode = htons((u_short)DATA);
		dp->th_block = htons((u_short)cl->block);
	}

	if (cl->state == ST_SEND) {
		/* See if it's time to give up on this turkey */
		if (cl->retries-- < 0) {
			++stats.timeout;
			lg(LOG_INFO, "%s.%d timeout",
			    inet_ntoa(cl->sin.sin_addr), cl->sin.sin_port);
			cl->proc = CP_DONE;
			return;
		}

		/* Send the block */
		if (sendto(s, (char *)cl->dp, cl->dpcc, 0,
		    (struct sockaddr *)&cl->sin, sizeof(cl->sin)) != cl->dpcc) {
			oerrno = errno;
			lg(LOG_ERR, "xmitfile: write: %m");
			nak(cl, oerrno + 100, NULL);
			return;
		}
		++stats.sent;

		if (debug > 1)
			lg(LOG_DEBUG, "%s.%d > %d",
			    inet_ntoa(cl->sin.sin_addr),
			    cl->sin.sin_port,
			    cl->block);

		/* Read ahead next block */
		read_ahead(&cl->ra, cl->file, cl->convert, cl->segsize);

		if (size == cl->segsize)
			cl->state = ST_WAITACK;
		else
			cl->state = ST_LASTACK;
		cl->wakeup = now + rexmtval;
	}
}

static void
recvfile(struct client *cl)
{
	int oerrno, size, opcode;
	struct tftphdr *tp;
	struct bf *b;
	volatile u_short block;

	/* won't be a packet if we're doing a timeout */
	/* XXX move timeout code to a different routine? */

	/* Initialize some stuff the first time through */
	if (cl->state == ST_INIT) {
		cl->block = 0;
		cl->state = ST_RECV;
		cl->wakeup = 0;
		cl->retries = WAIT_RETRIES;
		cl->dp = w_init(&cl->ra);
		ack(cl);
		++cl->block;
		return;
	}

	if (cl->state != ST_RECV) {
		lg(LOG_ERR, "not state ST_RECV (%d)", cl->state);
		abort();
	}

	tp = (struct tftphdr *)cl->buf;
	if (tp == NULL) {

		/* See if it's time to give up on this turkey */
		if (cl->retries-- < 0) {
			++stats.timeout;
			lg(LOG_INFO, "%s.%d timeout",
			    inet_ntoa(cl->sin.sin_addr), cl->sin.sin_port);
			cl->proc = CP_DONE;
			return;
		}
		return;
	}

	/* Process an input packet */
	opcode = ntohs((u_short)tp->th_opcode);

	/* Give up if the other side wants to */
	if (opcode == ERROR) {
		++stats.errorop;
		lg(LOG_INFO, "%s.%d ERROR",
		    inet_ntoa(cl->sin.sin_addr), cl->sin.sin_port);
		cl->proc = CP_DONE;
		return;
	}

	/* If he didn't send us data, ignore packet */
	/* XXX maybe he'll send a NAK if the packet gets dropped? */
	if (opcode != DATA) {
		/* XXX this probably can't happen */
		lg(LOG_DEBUG, "%s.%d recvfile: expected opcode %d",
		    inet_ntoa(cl->sin.sin_addr), cl->sin.sin_port,
		    opcode);
		return;
	}
	++stats.recv;

	block = ntohs((u_short)tp->th_block);
	if (block != cl->block) {
		++stats.wrongblock;
		/* XXX todo */
		/* Re-synchronize with the other side */
		(void) synchnet(&cl->ra);
		if (block == (cl->block - 1))
			ack(cl);		/* re-ACK last block */
		return;
	}

	/* XXX would like to avoid this copy */
	b = &cl->ra.bfs[cl->ra.current];
	if (cl->bufcc > sizeof(b->buf)) {
		lg(LOG_ERR, "buffer too big (%d > %d)",
		    cl->bufcc, sizeof(b->buf));
		abort();
	}
	memcpy(b->buf, cl->buf, cl->bufcc);
	b->counter = cl->bufcc;

	if (debug > 1)
		lg(LOG_DEBUG, "%s.%d < %d",
		    inet_ntoa(cl->sin.sin_addr),
		    cl->sin.sin_port,
		    cl->block);

	/* Read ahead next block */
	size = write_behind(&cl->ra, cl->file, cl->convert);
	if (size < 0) {
		oerrno = errno;
		lg(LOG_ERR, "write_behind: write: %m");
		nak(cl, oerrno + 100, NULL);
		return;
	}

	ack(cl);
	++cl->block;

	if (size == cl->segsize)
		cl->wakeup = now + TIMEOUT;
	else
		cl->proc = CP_DONE;
}

/* Process an incoming tftp packet */
static void
process(struct sockaddr_in *fromp, char *buf, int cc)
{
	struct tftphdr *tp;
	struct client *cl;
	u_short opcode;

	tp = (struct tftphdr *)buf;
	/* Look for old session */
	cl = findclient(fromp);

	opcode = ntohs(tp->th_opcode);

	/* We can receive an ACK to a sent NACK; eat it */
	if (cl == NULL && opcode == ACK)
		return;

	/* See if we need to reset him */
	if (cl != NULL && (opcode == RRQ || opcode == WRQ) &&
	    cl->proc == CP_SENDFILE && cl->state == ST_WAITACK) {
		++stats.badop;
		if (debug)
			lg(LOG_DEBUG, "%s.%d resetting",
			    inet_ntoa(cl->sin.sin_addr), cl->sin.sin_port);
		freeclient(cl);
		cl = NULL;
		/* Fall through and handle like a new client */
	}

	if (cl == NULL) {
		cl = newclient();
		if (cl == NULL)
			return;
		cl->sin = *fromp;
		cl->proc = CP_TFTP;
		if (debug)
			lg(LOG_DEBUG, "%s.%d starting",
			    inet_ntoa(cl->sin.sin_addr), cl->sin.sin_port);
	}
	cl->buf = buf;
	cl->bufcc = cc;

	/* Start up this client */
	runclient(cl);
}

/*
 * Scan client list for guys that need to retransmit. Do it backwards
 * in case anybody gets freed.
 */
static void
runtimer()
{
	struct client *cl, **clp;

	if (debug > 1)
		lg(LOG_DEBUG, "runtimer (clientlistcnt %d)", clientlistcnt);
	for (clp = clientlist + clientlistcnt - 1; clp >= clientlist; --clp) {
		cl = *clp;
		cl->buf = NULL;
		cl->bufcc = 0;
		if (cl->wakeup != 0 && cl->wakeup <= now) {
			if (debug)
				lg(LOG_DEBUG, "%s.%d wakeup",
				    inet_ntoa(cl->sin.sin_addr),
				    cl->sin.sin_port);
			runclient(cl);
		}
	}
}

/* Execute a client proc */
static void
runclient(struct client *cl)
{

	switch (cl->proc) {

	case CP_TFTP:
		tftp(cl);
		break;

	case CP_SENDFILE:
		xmitfile(cl);
		break;

	case CP_RECVFILE:
		recvfile(cl);
		break;

	case CP_OACK:
		recvoack(cl);
		break;

	default:
		lg(LOG_ERR, "bad client proc %d", cl->proc);
		finish(1);
	}
	if (cl->proc == CP_DONE)
		freeclient(cl);
}

/* Receive options ack ack */
static void
recvoack(struct client *cl)
{
	struct tftphdr *ap;

	ap = (struct tftphdr *)cl->buf;

	ap->th_opcode = ntohs((u_short)ap->th_opcode);
	ap->th_block = ntohs((u_short)ap->th_block);
	if (ap->th_opcode == ERROR)
		cl->proc = CP_DONE;
	else if (ap->th_opcode == ACK && ap->th_block == 0)
		cl->proc = CP_TFTP;
	/* XXX do we ever clean up? */
}

static struct errmsg {
	int	e_code;
	char	*e_msg;
} errmsgs[] = {
	{ EUNDEF,	"Undefined error code" },
	{ ENOTFOUND,	"File not found" },
	{ EACCESS,	"Access violation" },
	{ ENOSPACE,	"Disk full or allocation exceeded" },
	{ EBADOP,	"Illegal TFTP operation" },
	{ EBADID,	"Unknown transfer ID" },
	{ EEXISTS,	"File already exists" },
	{ ENOUSER,	"No such user" },
	{ 0,		NULL }
};

static char *
errtomsg(int error)
{
	struct errmsg *pe;
	static char buf[32];

	if (error == 0)
		return "success";
	for (pe = errmsgs; pe->e_msg != NULL; pe++)
		if (pe->e_code == error)
			return (pe->e_msg);
	sprintf(buf, "error %d", error);
	return (buf);
}

/*
 * Send a nak packet (error message).
 * Error code passed in is one of the
 * standard TFTP codes, or a UNIX errno
 * offset by 100.
 */
/* XXX estr not used yet */
static void
nak(struct client *cl, int error, const char *efmt, ...)
{
	int size;
	struct tftphdr *tp;
	struct errmsg *pe;
	va_list ap;
	char buf[PKTSIZE];
	char estr[80];

	memset(buf, 0, sizeof(buf));
	tp = (struct tftphdr *)buf;
	tp->th_opcode = htons((u_short)ERROR);
	tp->th_code = htons((u_short)error);
	estr[0] = '\0';
	if (efmt != NULL && efmt[0] != '\0') {
		va_start(ap, efmt);
		vsnprintf(estr, sizeof(estr), efmt, ap);
		va_end(ap);
	}
	for (pe = errmsgs; pe->e_code >= 0; pe++)
		if (pe->e_code == error)
			break;
	if (pe->e_code < 0) {
		pe->e_msg = strerror(error - 100);
		tp->th_code = EUNDEF;   /* set 'undef' errorcode */
	}

	if (estr[0] != '\0') {
		strcpy(tp->th_msg, estr);
		size = strlen(estr);
	} else {
		strcpy(tp->th_msg, pe->e_msg);
		size = strlen(pe->e_msg);
	}
	size += 4 + 1;
	if (sendto(s, buf, size, 0, (struct sockaddr *)&cl->sin,
	    sizeof(cl->sin)) != size)
		lg(LOG_ERR, "nak: %m");
	cl->proc = CP_DONE;
}

#ifdef HAVE_IPV6
/* translate IPv4 mapped IPv6 address to IPv4 address */
static void
unmappedaddr(struct sockaddr_in6 *sin6)
{
	struct sockaddr_in *sin4;
	u_int32_t addr;
	int port;

	if (sin6->sin6_family != AF_INET6 ||
	    !IN6_IS_ADDR_V4MAPPED(&sin6->sin6_addr))
		return;
	sin4 = (struct sockaddr_in *)sin6;
	addr = *(u_int32_t *)&sin6->sin6_addr.s6_addr[12];
	port = sin6->sin6_port;
	memset(sin4, 0, sizeof(struct sockaddr_in));
	sin4->sin_addr.s_addr = addr;
	sin4->sin_port = port;
	sin4->sin_family = AF_INET;
	sin4->sin_len = sizeof(struct sockaddr_in);
}
#endif

/*
 * Send an oack packet (option acknowledgement).
 */
static void
oack(struct client *cl)
{
	struct tftphdr *tp, *ap;
	int size, i, n;
	char *bp;
	char ackbuf[PKTSIZE];
	char *buf;

	buf = cl->buf;
	tp = (struct tftphdr *)buf;
	bp = buf + 2;
	size = PKTSIZE - 2;
	tp->th_opcode = htons((u_short)OACK);
	for (i = 0; cl->options[i].o_type != NULL; i++) {
		if (cl->options[i].o_request) {
			n = snprintf(bp, size, "%s%c%d", cl->options[i].o_type,
				     0, cl->options[i].o_reply);
			bp += n + 1;
			size -= n + 1;
			if (size < 0) {
				lg(LOG_ERR, "oack: buffer overflow");
				exit(1);
			}
		}
	}
	size = bp - buf;
	ap = (struct tftphdr *)ackbuf;

	if (sendto(s, buf, size, 0,
	    (struct sockaddr *)&cl->sin, sizeof(cl->sin)) != size) {
		lg(LOG_INFO, "oack: %m");
		cl->proc = CP_DONE;
		return;
	}
	cl->proc = CP_OACK;
}

/*
 * Send an ack packet.
 */
static void
ack(struct client *cl)
{
	size_t size;
	struct tftphdr *ap;
	char ackbuf[PKTSIZE];

	ap = (struct tftphdr *)ackbuf;
	memset(ap, 0, sizeof(*ap));
	ap->th_opcode = htons((u_short)ACK);
	ap->th_block = htons((u_short)cl->block);
	size = (char *)&ap->th_data - (char *)ap;

	if (sendto(s, ap, size, 0,
	    (struct sockaddr *)&cl->sin, sizeof(cl->sin)) != size) {
		lg(LOG_INFO, "ack: %m");
		cl->proc = CP_DONE;
		return;
	}
}

/* Exit with a syslog */
static __dead void
finish(int status)
{

	lg(LOG_DEBUG, "exiting");
	exit(status);
}

/* Signal handler for when something bad happens */
RETSIGTYPE
die(int signo)
{

	finish(1);
}

/* Signal handler to output statistics */
RETSIGTYPE
logstats(int signo)
{

	if (!candostats) {
		++dostats;
		return RETSIGVAL;
	}
	dostats = 0;
	lg(LOG_DEBUG, "logstats: %d clientlistcnt", clientlistcnt);
	lg(LOG_DEBUG, "logstats: %d retrans", stats.retrans);
	lg(LOG_DEBUG, "logstats: %d timeout", stats.timeout);
	lg(LOG_DEBUG, "logstats: %d badop", stats.badop);
	lg(LOG_DEBUG, "logstats: %d errorop", stats.errorop);
	lg(LOG_DEBUG, "logstats: %d badfilename", stats.badfilename);
	lg(LOG_DEBUG, "logstats: %d badfilemode", stats.badfilemode);
	lg(LOG_DEBUG, "logstats: %d wrongblock", stats.badfilemode);
	lg(LOG_DEBUG, "logstats: %d fileioerror", stats.fileioerror);
	lg(LOG_DEBUG, "logstats: %d sent", stats.sent);
	lg(LOG_DEBUG, "logstats: %d recv", stats.recv);
	lg(LOG_DEBUG, "logstats: %d maxclients", stats.maxclients);
	return RETSIGVAL;
}

/* Allocate a new client struct */
static struct client *
newclient()
{
	struct client *cl, **clp;
	u_int n, size;
	static int allocfailure = 0;

	if (clientlistcnt >= clientlistsize) {
		/* If we've every had an allocation failure, give up now */
		if (allocfailure)
			return (NULL);

		n = MORECLIENTS;

		/* Allocate array of pointers */
		if (clientlist == NULL) {
			size = n * sizeof(clientlist[0]);
			clientlist = (struct client **)malloc(size);
			if (clientlist == NULL) {
				lg(LOG_ERR, "newclient: list malloc: %m");
				finish(1);
			}
			clp = clientlist;
		} else {
			size = (clientlistsize + n) * sizeof(clientlist[0]);
			clientlist =
			    (struct client **)realloc(clientlist, size);
			if (clientlist == NULL) {
				lg(LOG_ERR, "newclient: list realloc: %m");
				++allocfailure;
				return (NULL);
			}
			clp = clientlist + clientlistsize;
		}

		size = n * sizeof(*cl);
		cl = (struct client *)malloc(size);
		if (cl == NULL) {
			lg(LOG_ERR, "newclient: client malloc: %m");
			++allocfailure;
			return (NULL);
		}
		memset(cl, 0, size);
		clientlistsize += n;
		while (n-- > 0)
			*clp++ = cl++;

	}
	cl = *(clientlist + clientlistcnt);
	++clientlistcnt;
	if (stats.maxclients < clientlistcnt)
		stats.maxclients = clientlistcnt;

	memcpy(cl->options, protooptions, sizeof(protooptions));
	cl->segsize = SEGSIZE;

	return (cl);
}

static struct client *
findclient(struct sockaddr_in *sinp)
{
	int n;
	struct client **clp;
	struct sockaddr_in *sinp2;

	n = clientlistcnt;
	clp = clientlist;
	while (n > 0 && (*clp)->proc != CP_FREE) {
		sinp2 = &(*clp)->sin;
		if (sinp->sin_port == sinp2->sin_port &&
		    sinp->sin_addr.s_addr == sinp2->sin_addr.s_addr)
			return (*clp);
		--n;
		++clp;
	}
	return (NULL);
}

static void
freeclient(struct client *cl)
{
	int n, size;
	struct client **clp;

	if (debug)
		lg(LOG_DEBUG, "%s.%d ending",
		    inet_ntoa(cl->sin.sin_addr), cl->sin.sin_port);

	if (cl->file != NULL)
		fclose(cl->file);
	memset(cl, 0, sizeof(*cl));
	for (n = 0, clp = clientlist; *clp != cl; ++n, ++clp)
		if (n >= clientlistcnt) {
			lg(LOG_ERR, "freeclient: can't find client");
			finish(1);
		}
	size = (clientlistcnt - (n + 1)) * sizeof(*clp);
	if (size > 0)
		memcpy(clientlist + n, clientlist + n + 1, size);
	--clientlistcnt;
	clp = clientlist + clientlistcnt;
	*clp = cl;
}

/* XXX */
/* Returns true if there is no dhcp config or the specified address
   is a client in the dhcpd config */
static int
goodclient(u_int32_t a)
{

	if (!havedhcpdconf || dhcpclient(a))
		return (1);

	return (0);
}
