/*
 * Copyright (c) 2008
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef lint
static const char rcsid[] =
    "@(#) $Id: util.c 131 2008-10-19 09:44:36Z leres $ (LBL)";
#endif

#include <sys/types.h>
#include <sys/time.h>

#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <syslog.h>

#include "util.h"

/* Forwards */
static const char *tsstr(void);
static void getts(struct timeval *);

static void
getts(struct timeval *tsp)
{

	if (gettimeofday(tsp, NULL) < 0) {
		/* Can't use lg() since it can indirectly call us */
		syslog(LOG_ERR, "getts: gettimeofday: %s", strerror(errno));
		exit(EX_OSERR);
	}
}

void
lg(int pri, const char *fmt, ...)
{
	va_list ap;

	if (debug) {
		fputs(tsstr(), stderr);
		putc(' ', stderr);
		va_start(ap, fmt);
		(void)vfprintf(stderr, fmt, ap);
		va_end(ap);
		putc('\n', stderr);
	}

	va_start(ap, fmt);
	vsyslog(pri, fmt, ap);
	va_end(ap);
}

static const char *
tsstr()
{
	struct timeval tv;
	static char buf[32];

	getts(&tv);
	snprintf(buf, sizeof(buf), "%ld.%06ld", tv.tv_sec, tv.tv_usec);
	return (buf);
}
